set search_path to public;

CREATE SCHEMA tree;
CREATE SCHEMA tree_util;




set search_path to tree_util;

CREATE TABLE tree_util.tree (
tree_name varchar primary key,
id serial unique not null,
description varchar 
);

CREATE TABLE tree_util.tree_node (
id bigserial primary key,
node_name varchar,
tree_id integer not null references tree_util.tree(id),
node_data_id bigint,
weight double precision DEFAULT 0.0,
internal_weight bigserial not null,
level integer,
parent_id integer references tree_util.tree_node(id) ON DELETE CASCADE
);


CREATE UNIQUE INDEX tree_name ON tree_util.tree_node ( tree_id,node_name);


CREATE TABLE tree_util.tree_node_metadata (
id bigint primary key references tree_util.tree_node(id) ON DELETE CASCADE,
--node_name varchar unique,
node_description varchar
);

set search_path to public;

--DROP TYPE tree.traversal_node CASCADE;

CREATE TYPE tree.tree AS (
tree_name varchar,
description varchar 
);


CREATE TYPE tree.tree_node AS (
	id bigint,
	node_name varchar,
	node_data_id bigint,
    weight double precision,
    internal_weight bigint,
    level integer,
    parent_id bigint
);


CREATE TYPE tree.traversal_node AS (
	id bigint,
	node_name varchar,
    node_data_id bigint,
    weight double precision,
    internal_weight bigint,
    level integer,
    parent_id bigint,
    traversal_index integer
);


CREATE TYPE tree.traversal_group AS (
    group_tag integer,
	node_name varchar,
	node_id bigint,
    node_data_id bigint,
    weight double precision,
    internal_weight bigint,
    level integer,
    parent_id bigint,
    node_traversal_index integer,
    traversal_index integer
);



CREATE TYPE tree.tree_node_metadata AS (
id bigint ,
--node_name varchar,
node_description varchar
);



CREATE FUNCTION tree_util.cmp_tree_nodes(tree.tree_node  , tree.tree_node )
RETURNS boolean AS $$
  SELECT $1.id = $2.id;
$$ IMMUTABLE STRICT LANGUAGE SQL;



CREATE OPERATOR  = (
  PROCEDURE =  tree_util.cmp_tree_nodes,
  LEFTARG = tree.tree_node,
  RIGHTARG = tree.tree_node
);



--CREATE FUNCTION tree_util.cmp2_tree_nodes(tree.tree_node  , tree.tree_node )
--RETURNS boolean AS $$
--  SELECT $1.id = $2.id 
--	and $1.node_name = $2.node_name 
--	and $1.node_data_id = $2.node_data_id
--	and $1.weight = $2.weight
--	and $1.internal_weight = $2.internal_weight
--	and $1.parent_id = $2.parent_id
--;
--$$ IMMUTABLE STRICT LANGUAGE SQL;

--CREATE OPERATOR  == (
--  PROCEDURE =  tree_util.cmp2_tree_nodes,
--  LEFTARG = tree.tree_node,
--  RIGHTARG = tree.tree_node
--);



CREATE FUNCTION tree_util.lt_level_tree_nodes(tree.tree_node  , tree.tree_node )
RETURNS boolean AS $$
  --SELECT COALESCE($1.level,-1) < COALESCE($2.level,-1);
  SELECT $1.level < $2.level;
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OPERATOR  << (
  PROCEDURE =  tree_util.lt_level_tree_nodes,
  LEFTARG = tree.tree_node,
  RIGHTARG = tree.tree_node
);

CREATE FUNCTION tree_util.gt_level_tree_nodes(tree.tree_node  , tree.tree_node )
RETURNS boolean AS $$
  --SELECT COALESCE($1.level,-1) < COALESCE($2.level,-1);
  SELECT $1.level > $2.level;
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OPERATOR  >> (
  PROCEDURE =  tree_util.gt_level_tree_nodes,
  LEFTARG = tree.tree_node,
  RIGHTARG = tree.tree_node
);

