
set search_path to public;

CREATE SCHEMA tree;
CREATE SCHEMA tree_util;




set search_path to tree_util;

CREATE TABLE tree_util.tree (
tree_name varchar primary key,
id serial unique not null,
description varchar 
);

CREATE TABLE tree_util.tree_node (
id bigserial primary key,
node_name varchar,
tree_id integer not null references tree_util.tree(id),
node_data_id bigint,
weight double precision DEFAULT 0.0,
internal_weight bigserial not null,
level integer,
parent_id integer references tree_util.tree_node(id) ON DELETE CASCADE
);


CREATE UNIQUE INDEX tree_name ON tree_util.tree_node ( tree_id,node_name);


CREATE TABLE tree_util.tree_node_metadata (
id bigint primary key references tree_util.tree_node(id) ON DELETE CASCADE,
--node_name varchar unique,
node_description varchar
);

set search_path to public;

--DROP TYPE tree.traversal_node CASCADE;

CREATE TYPE tree.tree AS (
tree_name varchar,
description varchar 
);


CREATE TYPE tree.tree_node AS (
	id bigint,
	node_name varchar,
	node_data_id bigint,
    weight double precision,
    internal_weight bigint,
    level integer,
    parent_id bigint
);


CREATE TYPE tree.traversal_node AS (
	id bigint,
	node_name varchar,
    node_data_id bigint,
    weight double precision,
    internal_weight bigint,
    level integer,
    parent_id bigint,
    traversal_index integer
);


CREATE TYPE tree.traversal_group AS (
    group_tag integer,
	node_name varchar,
	node_id bigint,
    node_data_id bigint,
    weight double precision,
    internal_weight bigint,
    level integer,
    parent_id bigint,
    node_traversal_index integer,
    traversal_index integer
);



CREATE TYPE tree.tree_node_metadata AS (
id bigint ,
--node_name varchar,
node_description varchar
);



CREATE FUNCTION tree_util.cmp_tree_nodes(tree.tree_node  , tree.tree_node )
RETURNS boolean AS $$
  SELECT $1.id = $2.id;
$$ IMMUTABLE STRICT LANGUAGE SQL;



CREATE OPERATOR  = (
  PROCEDURE =  tree_util.cmp_tree_nodes,
  LEFTARG = tree.tree_node,
  RIGHTARG = tree.tree_node
);



--CREATE FUNCTION tree_util.cmp2_tree_nodes(tree.tree_node  , tree.tree_node )
--RETURNS boolean AS $$
--  SELECT $1.id = $2.id 
--	and $1.node_name = $2.node_name 
--	and $1.node_data_id = $2.node_data_id
--	and $1.weight = $2.weight
--	and $1.internal_weight = $2.internal_weight
--	and $1.parent_id = $2.parent_id
--;
--$$ IMMUTABLE STRICT LANGUAGE SQL;

--CREATE OPERATOR  == (
--  PROCEDURE =  tree_util.cmp2_tree_nodes,
--  LEFTARG = tree.tree_node,
--  RIGHTARG = tree.tree_node
--);



CREATE FUNCTION tree_util.lt_level_tree_nodes(tree.tree_node  , tree.tree_node )
RETURNS boolean AS $$
  --SELECT COALESCE($1.level,-1) < COALESCE($2.level,-1);
  SELECT $1.level < $2.level;
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OPERATOR  << (
  PROCEDURE =  tree_util.lt_level_tree_nodes,
  LEFTARG = tree.tree_node,
  RIGHTARG = tree.tree_node
);

CREATE FUNCTION tree_util.gt_level_tree_nodes(tree.tree_node  , tree.tree_node )
RETURNS boolean AS $$
  --SELECT COALESCE($1.level,-1) < COALESCE($2.level,-1);
  SELECT $1.level > $2.level;
$$ IMMUTABLE STRICT LANGUAGE SQL;

CREATE OPERATOR  >> (
  PROCEDURE =  tree_util.gt_level_tree_nodes,
  LEFTARG = tree.tree_node,
  RIGHTARG = tree.tree_node
);

---------------------------------------------------------------------------
-- tree_util.get_node(tree_util.tree_node) :  tree.tree_node
---------------------------------------------------------------------------
 
CREATE OR REPLACE FUNCTION tree_util.unode_to_node(arg_node tree_util.tree_node) RETURNS  tree.tree_node AS $$
DECLARE
	tn tree.tree_node;
begin
    tn.id := arg_node.id;
    tn.node_name := arg_node.node_name;
    tn.node_data_id := arg_node.node_data_id;
    tn.weight := arg_node.weight;
   	tn.internal_weight := arg_node.internal_weight;
    tn.level := arg_node.level;
	tn.parent_id := arg_node.parent_id;
	return  tn;
END
$$ IMMUTABLE LANGUAGE plpgsql;
  

CREATE CAST (tree_util.tree_node AS tree.tree_node)
  WITH FUNCTION tree_util.unode_to_node (tree_util.tree_node);
  
  
CREATE OR REPLACE FUNCTION tree_util.traversal_node_to_node(arg_node tree.traversal_node) RETURNS  tree.tree_node AS $$
DECLARE
	tn tree.tree_node;
begin
    tn.id := arg_node.id;
    tn.node_name := arg_node.node_name;
    tn.node_data_id := arg_node.node_data_id;
    tn.weight := arg_node.weight;
   	tn.internal_weight := arg_node.internal_weight;
    tn.level := arg_node.level;
	tn.parent_id := arg_node.parent_id;
	return  tn;
END
$$ IMMUTABLE LANGUAGE plpgsql;
  

CREATE CAST (tree.traversal_node AS tree.tree_node)
  WITH FUNCTION tree_util.traversal_node_to_node (tree.traversal_node);
  
  
  

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 

CREATE OR REPLACE FUNCTION tree_util.unode_metadata_to_node_metadata(arg_metadata tree_util.tree_node_metadata) RETURNS  tree.tree_node_metadata AS $$
DECLARE
	ma tree.tree_node_metadata;
begin
    	ma.id := arg_metadata.id;
		--ma.node_name := arg_metadata.node_name;
		ma.node_description := arg_metadata.node_description;
		return  ma;
END
$$ IMMUTABLE LANGUAGE plpgsql;


CREATE CAST (tree_util.tree_node_metadata AS tree.tree_node_metadata)
  WITH FUNCTION tree_util.unode_metadata_to_node_metadata( tree_util.tree_node_metadata);
  
  

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
  
CREATE OR REPLACE FUNCTION tree_util.utree_to_tree(arg_tree tree_util.tree) RETURNS  tree.tree AS $$
DECLARE
	t tree.tree;
begin
	t.tree_name := arg_tree.tree_name;
	t.description := arg_tree.description;
	return  t;
END
$$ IMMUTABLE LANGUAGE plpgsql;

CREATE CAST (tree_util.tree AS tree.tree)
  WITH FUNCTION tree_util.utree_to_tree( tree_util.tree);
  

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 

CREATE OR REPLACE FUNCTION tree_util.get_traversal_node(arg_node tree_util.tree_node, arg_traversal_index bigint) RETURNS  tree.traversal_node AS $$
DECLARE
	tn tree.traversal_node;
begin
    	tn.id := arg_node.id;
    	tn.node_name := arg_node.node_name;
    	tn.traversal_index := arg_traversal_index;
    	tn.node_data_id := arg_node.node_data_id;
        tn.weight := arg_node.weight;
    	tn.internal_weight := arg_node.internal_weight;
    	tn.level := arg_node.level;
		tn.parent_id := arg_node.parent_id;
		return  tn;
END
$$ IMMUTABLE LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION tree_util.get_traversal_node(arg_node tree.tree_node, arg_traversal_index bigint) RETURNS  tree.traversal_node AS $$
DECLARE
	tn tree.traversal_node;
begin
    	tn.id := arg_node.id;
    	tn.node_name := arg_node.node_name;
    	tn.traversal_index := arg_traversal_index;
    	tn.node_data_id := arg_node.node_data_id;
        tn.weight := arg_node.weight;
    	tn.internal_weight := arg_node.internal_weight;
    	tn.level := arg_node.level;
		tn.parent_id := arg_node.parent_id;
		return  tn;
END
$$ IMMUTABLE LANGUAGE plpgsql;

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 

CREATE OR REPLACE FUNCTION tree_util.get_traversal_group(arg_node tree.tree_node, arg_traversal_index integer, arg_group_tag integer, arg_node_traversal_index integer) RETURNS  tree.traversal_group AS $$
DECLARE
	tn tree.traversal_group;
begin
		tn.group_tag := arg_group_tag;
    	tn.node_id := arg_node.id;
    	tn.node_traversal_index := arg_node_traversal_index;
    	tn.traversal_index := arg_traversal_index;
    	tn.node_data_id := arg_node.node_data_id;
        tn.weight := arg_node.weight;
    	tn.internal_weight := arg_node.internal_weight;
    	tn.level := arg_node.level;
		tn.parent_id := arg_node.parent_id;
		return  tn;
END
$$ IMMUTABLE LANGUAGE plpgsql;



---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 
CREATE OR REPLACE FUNCTION tree_util.get_tree_id(arg_tree_name varchar) RETURNS  integer AS $$
	SELECT id from tree_util.tree WHERE tree_name = $1;
$$ STABLE LANGUAGE sql;


---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 


 CREATE OR REPLACE FUNCTION  tree_util.get_root(arg_tree_id integer) RETURNS  tree.tree_node AS $$
   SELECT (tn.*)::tree.tree_node FROM tree_util.tree_node tn where level = 0 and tree_id = $1;
 $$ STABLE LANGUAGE sql;
 
---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 CREATE OR REPLACE FUNCTION  tree_util.get_root_id(arg_tree_id integer) RETURNS  bigint AS $$
   SELECT  id FROM tree_util.tree_node where level = 0 and tree_id = $1;
 $$ STABLE LANGUAGE sql;




---------------------------------------------------------------------------
---------------------------------------------------------------------------
---------------------------------------------------------------------------
---------------------------------------------------------------------------
---------------------------------------------------------------------------
-- tree
---------------------------------------------------------------------------
---------------------------------------------------------------------------
---------------------------------------------------------------------------
---------------------------------------------------------------------------
--------------------------------------------------------------------------- 





---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
CREATE OR REPLACE FUNCTION  tree.get_childs(arg_node_id bigint) RETURNS  setof  tree.tree_node AS $$
  	SELECT (tn.*)::tree.tree_node FROM tree_util.tree_node tn
    WHERE parent_id = $1
	ORDER BY tn.weight, tn.internal_weight
 $$ STABLE LANGUAGE sql;
 
 
 


---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
CREATE OR REPLACE FUNCTION  tree.get_nodes_for_data(arg_tree_name varchar, arg_node_data_id bigint) RETURNS  setof  tree.tree_node AS $$
  	SELECT (tn.*)::tree.tree_node FROM tree_util.tree_node tn
  	WHERE 
  	tree_id = tree_util.get_tree_id($1) AND node_data_id = $2
	ORDER BY tn.weight, tn.internal_weight
 $$ STABLE LANGUAGE sql; 
 
 
 
 
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.get_leaf_nodes(arg_tree_name varchar) RETURNS  setof  tree.tree_node AS $$
    SELECT (n.*)::tree.tree_node from tree_util.tree_node n  
    JOIN (
    (SELECT distinct id as family_id from tree_util.tree_node tn WHERE tn.tree_id = tree_util.get_tree_id($1))
    except
    (SELECT distinct parent_id  as family_id from  tree_util.tree_node)) f 
    ON ( n.id = f.family_id)
 $$ STABLE LANGUAGE sql;
 


---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
  
 
 CREATE OR REPLACE FUNCTION  tree.get_internal_nodes(arg_tree_name varchar) RETURNS  setof  tree.tree_node AS $$
    SELECT (n.*)::tree.tree_node  from tree_util.tree_node n  
    JOIN (SELECT distinct parent_id as family_id  from  tree_util.tree_node) f 
    ON (n.tree_id = tree_util.get_tree_id($1) and n.id = f.family_id)
 $$ STABLE LANGUAGE sql;
 
 
 
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 
 CREATE OR REPLACE FUNCTION  tree.get_root(arg_tree_name varchar) RETURNS  tree.tree_node AS $$
   SELECT  (n.*)::tree.tree_node FROM tree_util.tree_node n 
   where
   n.level = 0 
   and  n.tree_id = tree_util.get_tree_id($1);
 $$ STABLE LANGUAGE sql;
 
 
 CREATE OR REPLACE FUNCTION  tree.get_root_id(arg_tree_name varchar) RETURNS  bigint AS $$
 	SELECT (tree.get_root($1)).id;
 $$
 language sql stable;
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
  CREATE OR REPLACE FUNCTION  tree.get_parent(arg_node_id bigint) RETURNS  tree.tree_node AS $$
  	SELECT (tn.*)::tree.tree_node FROM tree_util.tree_node tn
  	JOIN tree_util.tree_node fm ON (tn.id = fm.parent_id)
	WHERE fm.id = $1 ;
 $$ STABLE LANGUAGE sql;

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.get_node(arg_node_id bigint) RETURNS  tree.tree_node AS $$
  	SELECT  (tn.*)::tree.tree_node FROM tree_util.tree_node tn
	WHERE tn.id = $1;
 $$ STABLE LANGUAGE sql;
 
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 

 
 CREATE OR REPLACE FUNCTION  tree.get_nodes(arg_tree_name varchar) RETURNS  setof  tree.tree_node AS $$
    SELECT (n.*)::tree.tree_node from tree_util.tree_node n  
    WHERE
      n.level is not null AND
      n.tree_id = tree_util.get_tree_id($1);
 $$ STABLE LANGUAGE sql;
 
---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 
 CREATE OR REPLACE FUNCTION  tree.get_all_nodes(arg_tree_name varchar) RETURNS  setof  tree.tree_node AS $$
  SELECT (n.*)::tree.tree_node from tree_util.tree_node n  
    WHERE
      n.tree_id = tree_util.get_tree_id($1);
 $$ STABLE LANGUAGE sql;
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.get_orphan_nodes(arg_tree_name varchar) RETURNS  setof  tree.tree_node AS $$
    SELECT (n.*)::tree.tree_node from tree_util.tree_node n
    WHERE
      n.level is null AND
      n.tree_id =  tree_util.get_tree_id($1);
 $$ STABLE LANGUAGE sql;
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION tree.get_node(arg_tree_name varchar, arg_node_name varchar)  RETURNS  tree.tree_node AS $$
 	select (n.*)::tree.tree_node  from tree_util.tree_node n
 	JOIN tree_util.tree t ON (n.tree_id = t.id) 
 	WHERE t.tree_name = $1 and 
 	n.node_name = $2;
 $$ STABLE LANGUAGE sql;
 
 

 
 
 CREATE OR REPLACE FUNCTION tree.get_node_id(arg_tree_name varchar, arg_node_name varchar)  RETURNS bigint AS $$
 
 	select n.id from tree_util.tree_node n
 	JOIN tree_util.tree t ON (n.tree_id = t.id) 
 	WHERE t.tree_name = $1 and 
 	n.node_name = $2;

 $$ STABLE LANGUAGE sql;
 
 
 

CREATE OR REPLACE FUNCTION tree.get_node_metadata(arg_node_id bigint)  RETURNS  tree.tree_node_metadata AS $$
 	select (mt.*)::tree.tree_node_metadata from tree_util.tree_node_metadata mt
 	WHERE id = $1;

 $$ STABLE LANGUAGE sql;
 



 
 
 CREATE OR REPLACE FUNCTION tree.get_named_nodes(arg_tree_name varchar)  RETURNS setof tree.tree_node AS $$
	select (n.*)::tree.tree_node  from tree_util.tree_node n
 	JOIN tree_util.tree t ON (n.tree_id = t.id) 
 	WHERE t.tree_name = $1 and
 	n.node_name is not null;

 $$ STABLE LANGUAGE sql;
 
 
---------------------------------------------------------------------------
--
---------------------------------------------------------------------------

 
 
CREATE OR REPLACE FUNCTION  tree.get_tree(arg_tree_name varchar) RETURNS  tree.tree AS $$
  	SELECT  (t.*)::tree.tree FROM tree_util.tree t
	WHERE t.tree_name = $1 ;
 $$ STABLE LANGUAGE sql;
 
 
 
 
---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.get_trees() RETURNS  setof  tree.tree AS $$
    SELECT (t.*)::tree.tree from tree_util.tree t;
 $$ STABLE LANGUAGE sql;
 
 
 
 
 
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.get_path(arg_node_id bigint,  arg_include_node boolean ) RETURNS  setof  tree.traversal_node AS $$
 DECLARE
 tn tree.tree_node;
 rep tree.traversal_node;
 nid bigint;
 i integer default 0;
 BEGIN
    nid := arg_node_id;
    
    if arg_include_node = true then
    	i := i + 1;
 		tn := tree.get_node(nid);
 	    rep := tree_util.get_traversal_node(tn,i);
 		return next rep;
 	end if;
 
 LOOP
 	tn := tree.get_parent(nid);
 	nid := tn.id;
	if nid is not null then	
		i := i + 1;
		rep := tree_util.get_traversal_node(tn,i);
		return next rep;
	end if; 	
    EXIT WHEN nid is null;  
 END LOOP;
 
 return;
 
 END
 $$ STABLE LANGUAGE plpgsql;
 
 
 CREATE OR REPLACE FUNCTION  tree.get_path(arg_node_id bigint,  arg_include_node boolean, arg_include_root boolean ) RETURNS  setof  tree.traversal_node AS $$
 DECLARE
 tn tree.tree_node;
 rep tree.traversal_node;
 nid bigint;
 i integer default 0;
 BEGIN
    nid := arg_node_id;
    
    if arg_include_node = true then
    	i := i + 1;
 		tn := tree.get_node(nid);
 	    rep := tree_util.get_traversal_node(tn,i);
 		return next rep;
 	end if;
 
 LOOP
 	tn := tree.get_parent(nid);
 	nid := tn.id;
 	if not arg_include_root and tn.parent_id is null then
 		nid := null;
 	end if;
	if nid is not null then	
		i := i + 1;
		rep := tree_util.get_traversal_node(tn,i);
		return next rep;
	end if; 	
    EXIT WHEN nid is null;  
 END LOOP;
 
 return;
 
 END
 $$ STABLE LANGUAGE plpgsql;
 
 
 --CREATE OR REPLACE FUNCTION  tree.get_anchestors(arg_node_id bigint,  arg_node_id bigint ) RETURNS  setof tree.tree_node AS $$
 --(SELECT id,node_name,node_data_id,weight, internal_weight,level,parent_id from tree.get_path($1,false,false) )  
 --INTERSECT 
 --(SELECT id,node_name,node_data_id,weight, internal_weight,level,parent_id from tree.get_path($2,false,false) );
 --$$ STABLE LANGUAGE sql;
 
 
  CREATE OR REPLACE FUNCTION  tree.get_anchestor(arg_node_id bigint,  arg_node_id bigint ) RETURNS  setof tree.tree_node AS $$
 (SELECT id,node_name,node_data_id,weight, internal_weight,level,parent_id from tree.get_path($1,true,false) )  
 INTERSECT 
 (SELECT id,node_name,node_data_id,weight, internal_weight,level,parent_id from tree.get_path($2,true,false) )
 order by level desc limit 1
 ;
 $$ STABLE LANGUAGE sql;
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------


CREATE OR REPLACE FUNCTION  tree.create_tree(arg_tree_name varchar, arg_tree_comment varchar) RETURNS integer AS $$
 DECLARE
	r tree.tree_node;
	p_tree_id integer;
	p_node_id bigint;
	p_root_level integer default 0;
	p_internal_weight bigint;
 BEGIN
 	if arg_tree_name is null then
 		RAISE NOTICE 'arg_tree_name IS MANTATORY';
 		return null;
   end if;
   
   p_tree_id := tree_util.get_tree_id(arg_tree_name);
   if p_tree_id is not null then
   	RAISE NOTICE 'tree with name % allredy exists', arg_tree_name;
	return null;
   end if;
   
   p_tree_id := nextval('tree_util.tree_id_seq');
   
   insert into tree_util.tree (tree_name,id,description) values (arg_tree_name,p_tree_id, arg_tree_comment);
   
  return p_tree_id; 
 END
 $$ LANGUAGE plpgsql;



--CREATE OR REPLACE FUNCTION  tree.create_tree( arg_tree_comment varchar) RETURNS integer AS $$
-- DECLARE
-- BEGIN
--   return tree.create_tree(null,arg_tree_comment);
-- END
-- $$ LANGUAGE plpgsql;




 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------


CREATE OR REPLACE FUNCTION  tree.rename_tree(arg_tree_current_name varchar, arg_tree_new_name varchar) RETURNS boolean AS $$
 DECLARE
 	rec_tree_old tree_util.tree;
 	rec_tree_new tree_util.tree;
 	p_new_id integer;
 	p_old_id integer;
 BEGIN
 	if arg_tree_current_name is null then
 		RAISE NOTICE 'arg_tree_current_name IS MANTATORY';
 		return false;
   end if; 
   if arg_tree_new_name is null then
 		RAISE NOTICE 'arg_tree_new_name IS MANTATORY';
 		return false;
   end if; 
   
  
	SELECT INTO rec_tree_old * FROM tree_util.tree  WHERE tree_name = arg_tree_current_name ;
    if not found then
   	 RAISE NOTICE 'tree with name % is not exists', arg_tree_current_name;
   	 return false;
    end if;
    
    SELECT INTO rec_tree_new * FROM tree_util.tree WHERE tree_name = arg_tree_new_name ;
    if found then
   	 RAISE NOTICE 'tree with name % allredy exists', arg_tree_new_name;
   	 return false;
    end if;

	p_old_id := rec_tree_old.id;
    p_new_id := nextval('tree_util.tree_id_seq');
	insert into tree_util.tree (tree_name, id, description) values (arg_tree_new_name, p_new_id, rec_tree_old.description );
   	
   	update tree_util.tree_node set tree_id = p_new_id WHERE tree_id = p_old_id;
   	delete from tree_util.tree where id = p_old_id;
   
  	--return arg_tree_new_id; 
  	return true;
 END
 $$ LANGUAGE plpgsql;




 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION  tree.update_tree_description(arg_tree_name varchar, arg_desc varchar) RETURNS boolean AS $$
 DECLARE
 	rec_tree tree_util.tree;
 BEGIN
 	if arg_tree_name is null then
 		RAISE NOTICE 'arg_tree_name IS MANTATORY';
 		return false;
   end if; 
   
    SELECT INTO rec_tree * FROM tree_util.tree  WHERE tree_name = arg_tree_name ;
    if not found then
   	 RAISE NOTICE 'tree with name % is not exists', arg_tree_name;
   	 return null;
    end if;
   
   	update tree_util.tree set description = arg_desc  WHERE id = rec_tree.id;
   
  	return true; 
 END
 $$ LANGUAGE plpgsql;
 
 
 
 
 
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 
 CREATE OR REPLACE FUNCTION  tree.delete_tree(arg_tree_name varchar) RETURNS  boolean AS $$
 DECLARE
  rec_node tree.tree_node;
  tmp_b boolean;
  rec RECORD;
  p_tree_id integer;
  te tree_util.tree;
 BEGIN
 
  		if (arg_tree_name is null) then
 			RAISE NOTICE 'arg_tree_name IS MANTATORY';
 			return false;
 		end if;
 
 		SELECT into te * from tree_util.tree WHERE tree_name = arg_tree_name;
 		if not found then
 			RAISE NOTICE 'TREE % NOT FOUND',arg_tree_name;
 			return false;
 		end if;
 		p_tree_id := te.id;
 		
 		rec_node := tree.get_root(arg_tree_name);
 		if rec_node.id is not null then
		 	tmp_b := tree.delete_subtree(rec_node.id);
		end if;
		
		delete from tree_util.tree WHERE id  = p_tree_id;
		return true;
 END
 $$ LANGUAGE plpgsql;
 
 
  

 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION  tree.add_orphan_node(arg_tree_name varchar, arg_node_data bigint, arg_node_weight double precision) RETURNS tree.tree_node AS $$
 DECLARE
	rec1 RECORD;
	r tree_util.tree_node;
	p_node_id bigint;
	p_level integer;
	p_tree_id integer;
	p_node_weight double precision DEFAULT 0.0;
	p_internal_weight bigint;
	rep tree.tree_node;

 BEGIN
    if (arg_tree_name is null) then
 	 	RAISE NOTICE 'arg_tree_name IS MANTATORY';
 	 	return null;
 	 end if;
 	 
	p_tree_id := tree_util.get_tree_id(arg_tree_name);
    if p_tree_id is  null then
   	 RAISE NOTICE 'tree with name % is not exists', arg_tree_name;
	 return null;
    end if;

	if (arg_node_weight is not null) then
		p_node_weight := arg_node_weight;
	end if;
	
   	
   p_node_id := nextval('tree_util.tree_node_id_seq');
   p_internal_weight := nextval('tree_util.tree_node_internal_weight_seq');
   r.id := p_node_id;
   r.node_data_id := arg_node_data;
   r.weight := p_node_weight;
   r.internal_weight := p_internal_weight;
   r.tree_id := p_tree_id;
   
   insert into tree_util.tree_node ( id , node_data_id , weight ,  internal_weight, tree_id)  
   values (r.id, r.node_data_id, r.weight,  r.internal_weight, r.tree_id  );

   rep :=  tree_util.unode_to_node(r);
   return rep;
 END
 $$ VOLATILE LANGUAGE plpgsql;

 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION  tree.add_named_orphan_node(arg_tree_name varchar, arg_node_name varchar, arg_node_data bigint, arg_node_weight double precision) RETURNS tree.tree_node AS $$
 DECLARE
 r tree.tree_node;
 tmp boolean;
 BEGIN
 	r := tree.add_orphan_node(arg_tree_name  ,arg_node_data , arg_node_weight);
 	if (r.id is null) then
 		return r;
 	end if;
 	
 	tmp := tree.set_node_name(r.id , arg_node_name);
 	if (tmp = false) then
 		RAISE EXCEPTION 'UNEXPECTED ERROR CANOT SET NAME TO NODE';
 	end if;
 
 	return r;
 END
 $$ VOLATILE LANGUAGE plpgsql;


 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------


CREATE OR REPLACE FUNCTION  tree.add_orphan_node(arg_tree_name varchar, arg_node_data bigint) RETURNS tree.tree_node AS $$
 DECLARE
    r tree.tree_node;
	def_weight double precision DEFAULT 0.0;
 BEGIN
     r:= tree.add_orphan_node(arg_tree_name, arg_node_data, def_weight);
     return r;
 END
 $$ VOLATILE LANGUAGE plpgsql;

 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION  tree.add_named_orphan_node(arg_tree_name varchar, arg_node_name varchar, arg_node_data bigint) RETURNS tree.tree_node AS $$
 DECLARE
 r tree.tree_node;
 tmp boolean;
BEGIN
 	r := tree.add_orphan_node(arg_tree_name ,arg_node_data);
 	if (r.id is null) then
 		return r;
 	end if;
 	
 	tmp := tree.set_node_name(r.id,arg_node_name);
 	if (tmp = false) then
 		RAISE EXCEPTION 'UNEXPECTED ERROR CANOT SET NAME TO NODE';
 	end if;
 	return r;
 END
 $$ VOLATILE LANGUAGE plpgsql;

 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 
 CREATE OR REPLACE FUNCTION  tree.set_root_node(arg_node bigint) RETURNS boolean AS $$
 DECLARE
 	rec1 RECORD;
 BEGIN
 	if (arg_node is null) then
 		RAISE NOTICE 'arg_node IS MANTATORY';
 	end if;
 
   SELECT INTO rec1 tn.level, tn.tree_id, tn.parent_id  
   FROM tree_util.tree_node tn WHERE tn.id = arg_node; 
   if not found then
   		RAISE NOTICE 'tree_node % not found ', arg_node;
   		return false;
   end if;

   if tree_util.get_root_id(rec1.tree_id) is not null then
   		RAISE NOTICE 'TREE HAS ALLREDY ROOT NODE';
   	    return false;
   end if;
   
   if rec1.parent_id is not null then
   		RAISE NOTICE 'NODE % IS NOT ORPHAN ',arg_node ;
   		return false;
   end if;
   
   if rec1.level is not null then
   		RAISE NOTICE 'NODE % IS NOT ORPHAN ',arg_node ;
   		return false;
   end if;
	
	update tree_util.tree_node SET level = 0  WHERE id = arg_node;
	
	return true;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 
CREATE OR REPLACE FUNCTION  tree.set_root_node(arg_tree_name varchar, arg_node_name varchar) RETURNS boolean AS $$
 DECLARE
 r tree.tree_node;
 rep boolean;
 tmp boolean;
BEGIN
	r := tree.get_node(arg_tree_name, arg_node_name);
	if (r.id is null) then
		RAISE NOTICE 'CANOT FOUND NODE WITH NAME %',arg_node_name;
		return false;
	end if;
 	rep := tree.set_root_node(r.id);
	return rep; 	
 END
 $$ VOLATILE LANGUAGE plpgsql;

 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.add_root_node(arg_tree_name varchar, arg_node_data bigint, arg_weight double precision ) RETURNS tree.tree_node AS $$
 DECLARE
    p_tree_id integer;
    s boolean;
    r tree.tree_node;
    rec RECORD;
 BEGIN
  	 if (arg_tree_name is null) then
 	 	RAISE NOTICE 'arg_tree_name IS MANTATORY';
 	 	return null;
 	 end if;
     r := tree.add_orphan_node(arg_tree_name, arg_node_data, arg_weight);
     if (r.id is null) then
     	return null;
     end if;
     s := tree.set_root_node(r.id);
     if (s = false) then
     	RAISE EXCEPTION 'UNEXPECTED ERROR: CANOT SET ROOT NODE';
     end if;
     
     return r;
 END
 $$ VOLATILE LANGUAGE plpgsql;


 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 

 CREATE OR REPLACE FUNCTION  tree.add_root_node(arg_tree_name varchar ) RETURNS tree.tree_node AS $$
 DECLARE
  r tree.tree_node;
 BEGIN
    r := tree.add_root_node(arg_tree_name,null,null);
    return r;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 

 CREATE OR REPLACE FUNCTION  tree.add_root_node(arg_tree_name varchar, arg_node_data bigint ) RETURNS tree.tree_node AS $$
 DECLARE
  r tree.tree_node;
 BEGIN
    r := tree.add_root_node(arg_tree_name,arg_node_data,null);
    return r;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 
 
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 
 
 CREATE OR REPLACE FUNCTION  tree.update_node_data(arg_node bigint, arg_node_data_id bigint) RETURNS boolean AS $$
 DECLARE
 	rec1 RECORD;
 BEGIN
	if (arg_node is null) then
 		RAISE NOTICE 'arg_node IS MANTATORY';
 	end if;
 	
 	update tree_util.tree_node set node_data_id = arg_node_data_id WHERE id = arg_node;
	if found then
		return true;
	end if;
	
	return false;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 
 CREATE OR REPLACE FUNCTION  tree.unset_root_node(arg_node bigint) RETURNS boolean AS $$
 DECLARE
 	rec1 RECORD;
 BEGIN
 	if (arg_node is null) then
 		RAISE NOTICE 'arg_node IS MANTATORY';
 	end if;
 
   SELECT INTO rec1 tn.level, tn.tree_id, tn.parent_id  
   FROM tree_util.tree_node tn WHERE tn.id = arg_node; 
   if not found then
   		RAISE NOTICE 'NODE % NOT FOUND',arg_node;
   		return false;
   end if;

   if rec1.parent_id is not null then
   		RAISE NOTICE 'NODE % IS NOT ROOT',arg_node;
   		return false;
   end if;
   
   if rec1.level != 0 then
   		RAISE NOTICE 'NODE % IS NOT ROOT',arg_node;
   		return false;
   end if;
	
	update tree_util.tree_node SET level = null  WHERE id = arg_node;
	
	return true;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 


 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------

 
CREATE OR REPLACE FUNCTION  tree.set_parentship(arg_parent_node bigint, arg_child_node bigint) RETURNS boolean AS $$
 DECLARE
	rec1 RECORD;
	r tree.tree_node;
	p_node_id bigint;
	p_level integer;
	p_tree_id integer;
	p_tree_id_p integer;
	p_tree_id_c integer;
	p_internal_weight bigint;
	
 BEGIN
 
 	if (arg_parent_node is null) then
 		RAISE NOTICE 'arg_parent_node IS MANTATORY';
 	end if;
   	if (arg_child_node is null) then
 		RAISE NOTICE 'arg_child_node IS MANTATORY';
 	end if;
 	
   --stixia parent_node kai tis ikogenias pou ine parent
   SELECT INTO rec1 tn.level, tn.tree_id  FROM tree_util.tree_node tn WHERE tn.id = arg_parent_node; 
   	if not found then
   		RAISE NOTICE 'CANOT FOUND PARENT NODE %',arg_parent_node ;
   		return false;
   	end if;
   
   p_level := rec1.level;
   p_tree_id_p := rec1.tree_id;
   
   --if p_level is null then
   --	return false;
   --end if;
   
   --stixia child_node
   SELECT INTO rec1  tn.level, tn.tree_id  FROM tree_util.tree_node tn WHERE tn.id = arg_child_node; 
   	if not found then
   		RAISE NOTICE 'CANOT FOUND CHILD NODE %',arg_child_node ;
   		return false;
   end if;
   p_tree_id_c := rec1.tree_id;
   
   if rec1.level is not null  then
   	RAISE NOTICE 'CHILD NODE % IS NOT ORPHAN  ',arg_child_node ;
   	return false;
   end if;
   
   if p_tree_id_c <> p_tree_id_p then
   	RAISE NOTICE 'NODES % % IS NOT IN SAME TREE',arg_parent_node, arg_child_node ;
   	return false;
   end if;
   
      
    if p_level is null then
		update tree_util.tree_node SET  parent_id = arg_parent_node WHERE id = arg_child_node;
    else
		p_level = p_level + 1;
   		update tree_util.tree_node SET level = p_level, parent_id = arg_parent_node WHERE id = arg_child_node;
    end if;
   
  
   return true;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 
 
 
 
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------

 
CREATE OR REPLACE FUNCTION  tree.set_parentship(arg_tree_name varchar, arg_parent_node_name varchar , arg_child_node_name varchar) RETURNS boolean AS $$
 DECLARE
 r1	 bigint;
 r2	 bigint;
 rep boolean;
BEGIN
	r1 := tree.get_node_id(arg_tree_name, arg_parent_node_name);
	if (r1 is null) then
		RAISE NOTICE 'CANOT FOUND NODE WITH NAME %',arg_node_name;
		return false;
	end if;

	r2 := tree.get_node_id(arg_tree_name, arg_child_node_name);
	if (r1 is null) then
		RAISE NOTICE 'CANOT FOUND NODE WITH NAME %',arg_node_name;
		return false;
	end if;
 	
 	rep := tree.set_parentship(r1,r2);
	return rep; 	
 END
 $$ VOLATILE LANGUAGE plpgsql;
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.make_orphan_node(arg_node bigint) RETURNS boolean AS $$
 DECLARE
 	rec1 RECORD;
 BEGIN
 
   if (arg_node is null) then
 		RAISE NOTICE 'arg_node IS MANTATORY';
   end if;
 
   SELECT INTO rec1 tn.level, tn.tree_id, tn.parent_id  
   FROM tree_util.tree_node tn WHERE tn.id = arg_node; 
   if not found then
        RAISE NOTICE 'arg_node % NOT FOUND',arg_node ;
   		return false;
   end if;

	
	update tree_util.tree_node SET level = null, parent_id = null  WHERE id = arg_node;
	
	return true;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 

 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------


CREATE OR REPLACE FUNCTION  tree.add_node(arg_parent_node bigint, arg_node_data bigint, arg_node_weight double precision) RETURNS tree.tree_node AS $$
 DECLARE
	rec1 RECORD;
	r tree_util.tree_node;
	rep tree.tree_node;
	p_node_id bigint;
	p_level integer;
	p_tree_id integer;
	p_internal_weight bigint;
	p_weight double precision DEFAULT 0.0;
	
 BEGIN
   
   --stixia parent_node kai tis ikogenias pou ine parent
   SELECT INTO rec1 tn.level, tn.tree_id  FROM tree_util.tree_node tn WHERE tn.id = arg_parent_node; 
   	if not found then
   		RAISE NOTICE 'NODE % NOT FOUND ',arg_parent_node;
   		return null;
   	end if;
   
   p_level := rec1.level;
   p_tree_id := rec1.tree_id;
      
   -- level kenouriou komvou
   p_level = p_level + 1;
   
   if (arg_node_weight is not null) then
   	p_weight := arg_node_weight;
   end if;
   
   p_node_id := nextval('tree_util.tree_node_id_seq');
   p_internal_weight := nextval('tree_util.tree_node_internal_weight_seq');
   r.id := p_node_id;
   r.node_data_id := arg_node_data;
   r.weight := p_weight;
   r.level := p_level;
   r.internal_weight := p_internal_weight;
   r.tree_id := p_tree_id;
   insert into tree_util.tree_node ( id , node_data_id , weight , level, internal_weight, tree_id,  parent_id)  
   values (r.id, r.node_data_id, r.weight, r.level, r.internal_weight, r.tree_id , arg_parent_node );
  
   rep :=  tree_util.unode_to_node(r);
   return rep;
 END
 $$ VOLATILE LANGUAGE plpgsql;



 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------




CREATE OR REPLACE FUNCTION  tree.add_node(arg_parent_node bigint, arg_node_data bigint) RETURNS tree.tree_node AS $$
 DECLARE
    r tree.tree_node;
 BEGIN
     r:= tree.add_node(arg_parent_node, arg_node_data, null);
     return r;
 END
 $$ LANGUAGE plpgsql;
 
 
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 
 
 
 CREATE OR REPLACE FUNCTION  tree.swap_childs_order(arg_node_id_1 bigint, arg_node_id_2 bigint) RETURNS  boolean AS $$
 DECLARE
  rec_node_1 tree.tree_node;
  rec_node_2 tree.tree_node;
  tmp_i1 bigint;
  tmp_d1 double precision;
  tmp_i2 bigint;
  tmp_d2 double precision;
 BEGIN
	rec_node_1 := tree.get_node(arg_node_id_1); 
	rec_node_2 := tree.get_node(arg_node_id_2);
	if (rec_node_1 is null or rec_node_2 is null) then
		return false;
	end if;

	if (rec_node_1.parent_id is null or  rec_node_2.parent_id is null) then
		return false;
	end if;
	
	if (rec_node_1.parent_id != rec_node_2.parent_id) then
		return false;
	end if;
	
	tmp_i1 := rec_node_1.internal_weight;
	tmp_d1 := rec_node_1.weight;
	tmp_i2 := rec_node_2.internal_weight;
	tmp_d2 := rec_node_2.weight;
	
	UPDATE tree_util.tree_node SET  internal_weight = tmp_i2 , weight = tmp_d2 WHERE id = arg_node_id_1;
	UPDATE tree_util.tree_node SET  internal_weight = tmp_i1 , weight = tmp_d1 WHERE id = arg_node_id_2;
	
 return true;   
 END
 $$ LANGUAGE plpgsql;
 
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 
 
 CREATE OR REPLACE FUNCTION  tree.delete_node(arg_node_id bigint) RETURNS  boolean AS $$
 DECLARE
 p_parent bigint;
 rec_node_1 tree.tree_node;
 rec_node_2 tree.tree_node;
 i bigint DEFAULT 0;
 BEGIN
 
 if (arg_node_id is null) then
 		RAISE NOTICE 'arg_node IS MANTATORY';
 		return false;
 end if;
   
 SELECT INTO rec_node_1 * FROM tree.get_parent(arg_node_id);
 p_parent := rec_node_1.id;
 
 IF p_parent IS NULL THEN
 	return false;
 END IF;

 FOR rec_node_1 IN 
 	SELECT * FROM tree.get_childs(p_parent)
 	LOOP 
 	i := i + 1;
 	if rec_node_1.id = arg_node_id then
 		FOR rec_node_2 IN	
 		SELECT * FROM tree.get_childs(arg_node_id)
		LOOP
			i := i+1;
			UPDATE tree_util.tree_node SET level = level - 1, parent_id = p_parent, internal_weight = i WHERE id = rec_node_2.id;
		END LOOP;
	else 
		UPDATE tree_util.tree_node SET internal_weight = i WHERE id = rec_node_1.id;
 	end if;
 	
 	END LOOP;
 	
	DELETE FROM  tree_util.tree_node WHERE id  =  arg_node_id;

 return true;   
 END
 $$ LANGUAGE plpgsql;
 



 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
  
 
 
 CREATE OR REPLACE FUNCTION  tree.delete_subtree(arg_node_id bigint) RETURNS  boolean AS $$
 DECLARE
 BEGIN
 	
 	 if (arg_node_id is null) then
 		RAISE NOTICE 'arg_node IS MANTATORY';
 		return false;
 	end if;
 
 	if arg_node_id is null then
 		return false;
 	end if;
 	
		DELETE FROM  tree_util.tree_node WHERE id =  arg_node_id;
		if found then
 			return true;
 		end if;
 		
 		return false;  
 END
 $$ LANGUAGE plpgsql;
 
 
 
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION tree.set_node_name(arg_node_id bigint, arg_node_name varchar)  RETURNS  boolean AS $$
 DECLARE
 	rec tree_util.tree_node_metadata;
	rec_node tree.tree_node;
 BEGIN
 	
 	update tree_util.tree_node set node_name = arg_node_name where id = arg_node_id;
	if found then
		return true;
	end if;
	
 	return false;
	
 END
 $$ LANGUAGE plpgsql;
 
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION tree.set_node_description(arg_node_id bigint, arg_node_description varchar)  RETURNS  boolean AS $$
 DECLARE
 	rec tree_util.tree_node_metadata;
	rec_node tree.tree_node;
 BEGIN
 	rec_node := tree.get_node(arg_node_id);
 	if (rec_node.id is null) then
 		RAISE NOTICE 'NODE % NOT FOUND',arg_node_id;
 		return false;
 	end if;
 	
 	select into rec * from tree.tree_node_metadata where id = arg_node_id;
 	if (rec.id is not null) then
 		update tree_util.tree_node_metadata set node_description = arg_node_description WHERE id = arg_node_id;
 	else
 		insert into tree_util.tree_node_metadata (id,node_description) values (arg_node_id, arg_node_description);
 	end if;
 	
 	return true;
	
 END
 $$ LANGUAGE plpgsql;
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION tree.set_node_metadata(arg_node_id bigint,  arg_node_description varchar)  RETURNS  boolean AS $$
 DECLARE
 	rec tree_util.tree_node_metadata;
	rec_node tree.tree_node;
 BEGIN
 	rec_node := tree.get_node(arg_node_id);
 	if (rec_node.id is null) then
 		RAISE NOTICE 'NODE % NOT FOUND',arg_node_id;
 		return false;
 	end if;
 	
 	select into rec * from tree_util.tree_node_metadata where id = arg_node_id;
 	if (rec.id is not null) then
 		update tree_util.tree_node_metadata  SET
 		    node_description = arg_node_description 
 		WHERE id = arg_node_id;
 	else
 		insert into tree_util.tree_node_metadata (id,node_description) values (arg_node_id, arg_node_description);
 	end if;
 	
 	return true;
	
 END
 $$ LANGUAGE plpgsql;
 
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION tree.remove_node_name(arg_node_id bigint)  RETURNS  boolean AS $$
 DECLARE
 BEGIN
 
	update tree_util.tree_node set node_name = null WHERE id = arg_node_id;
	if found then
		return true;
	end if;
	return false;
	
 END
 $$ LANGUAGE plpgsql;
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION tree.remove_node_metadata(arg_node_id bigint)  RETURNS  boolean AS $$
 DECLARE
 BEGIN
 
	delete from tree_util.tree_node_metadata  WHERE id = arg_node_id;
	if found then
		return true;
	end if;
	return false;
	
 END
 $$ LANGUAGE plpgsql;
 
 


 ---------------
 -- TRIGGER
 ---------------
 


CREATE OR REPLACE FUNCTION tree_util.trigger_level_set() RETURNS trigger AS $$
DECLARE
	rec_node tree_util.tree_node;
	rec RECORD;
	p_tree_id integer;
	p_parent_id bigint;
	p_level_parent integer;
	p_level integer;
BEGIN
	  p_tree_id := NEW.tree_id;
	        
      FOR rec_node IN
      SELECT * FROM tree_util.tree_node WHERE tree_id = p_tree_id AND level is null AND parent_id is not null
      LOOP
      	p_parent_id := rec_node.parent_id;
		SELECT INTO rec  level FROM tree_util.tree_node WHERE id = p_parent_id;
		p_level_parent := rec.level;
		if p_level_parent is not null then
			p_level := p_level_parent +1;
			update tree_util.tree_node set level = p_level WHERE id = rec_node.id;
		end if;
      END LOOP;    
          
      RETURN NEW;
END
$$ LANGUAGE plpgsql;
 
 
set search_path to tree_util;
 
CREATE TRIGGER tree_node_after AFTER INSERT OR UPDATE ON  "tree_node"
FOR EACH ROW EXECUTE PROCEDURE tree_util.trigger_level_set();

set search_path to public; 


 ---------------------------------------------------------------------------
 --tree_util.traverse_node
 ---------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION tree_util.traverse_node(arg_node tree.tree_node, arg_max_level integer) RETURNS  setof  tree.tree_node AS $$
DECLARE
	id1 tree.tree_node ;
	id2 tree.tree_node ;
begin
    return next arg_node;
    FOR id1 IN
    SELECT * FROM tree.get_childs(arg_node.id)
    LOOP
    	IF arg_max_level < 0 OR id1.level <= arg_max_level THEN
         FOR id2 IN
         SELECT * FROM tree_util.traverse_node(id1,arg_max_level)
         LOOP
            return next id2;
         END LOOP;
        END IF;
    END LOOP;
END
$$ STABLE LANGUAGE plpgsql;


---------------------------------------------------------------------------
---------------------------------------------------------------------------
-- tree
---------------------------------------------------------------------------
---------------------------------------------------------------------------



---------------------------------------------------------------------------
--
---------------------------------------------------------------------------




CREATE OR REPLACE FUNCTION tree_util.traverse_node(arg_node_id bigint, arg_max_level integer) RETURNS  setof  tree.traversal_node AS $$
DECLARE
	rec tree.tree_node ;
	tn tree.traversal_node;
	i integer DEFAULT 0;
begin
    FOR rec IN
    SELECT * FROM tree_util.traverse_node(tree.get_node(arg_node_id), arg_max_level)
    LOOP
    	i := i + 1;
    	tn := tree_util.get_traversal_node(rec,i);
		return next tn;
    END LOOP;
    return;
END
$$ STABLE LANGUAGE plpgsql;




CREATE OR REPLACE FUNCTION tree.traverse_node(arg_node_id bigint, arg_max_level integer) RETURNS  setof  tree.traversal_node AS $$
	SELECT * FROM tree_util.traverse_node($1,$2);
$$ STABLE LANGUAGE sql;

CREATE OR REPLACE FUNCTION tree.traverse_node(arg_node_id bigint) RETURNS  setof  tree.traversal_node AS $$
	SELECT * FROM tree_util.traverse_node($1,-1);
$$ STABLE LANGUAGE sql;

 
 CREATE OR REPLACE FUNCTION  tree.traverse_tree(arg_tree_name varchar) RETURNS  setof  tree.traversal_node AS $$
 SELECT * FROM tree_util.traverse_node(tree.get_root_id($1),-1);
 $$ STABLE LANGUAGE sql;

 
 CREATE OR REPLACE FUNCTION  tree.traverse_tree(arg_tree_name varchar, arg_max_level integer) RETURNS  setof  tree.traversal_node AS $$
 SELECT * FROM tree_util.traverse_node(tree.get_root_id($1),$2);
 $$ STABLE LANGUAGE sql;
 
 
 
 
 
 
 
 
 
 
 
 
 


---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree_util.traverse_tree_level(arg_tree_name varchar) RETURNS  setof  tree.traversal_node AS $$
 DECLARE
 rec RECORD;
 rep tree.traversal_node;
 i integer DEFAULT 0;
 p_t_id integer; 
 BEGIN
 
   if arg_tree_name is null then
 		RAISE NOTICE 'arg_tree_name IS MANTATORY';
 		return ;
   end if; 
   
   p_t_id := tree_util.get_tree_id(arg_tree_name);
   if p_t_id is null then
   	RAISE NOTICE 'tree with name % not exists', arg_tree_name;
	return;
   end if;
   
 
   FOR rec IN
      SELECT * FROM
	  (SELECT 
	  n.*, 
      (tree.get_parent(n.id)) as parent
      FROM tree_util.tree_node n  
      WHERE
      n.tree_id = p_t_id ) as s
      order by level, (parent).weight, (parent).internal_weight, weight, internal_weight
    LOOP
    	i := i+1;
    	rep := tree_util.get_traversal_node((rec.id, rec.node_name, rec.node_data_id, rec.weight, rec.internal_weight, rec.level, rec.parent_id )::tree.tree_node,i);
		return next rep;    	
    END LOOP;
    return;
   
 END
 $$ STABLE LANGUAGE plpgsql;
 
 
 
  CREATE OR REPLACE FUNCTION  tree_util.traverse_tree_level(arg_tree_name varchar, arg_max_level integer) RETURNS  setof  tree.traversal_node AS $$
 DECLARE
 rec RECORD;
 rep tree.traversal_node;
 i integer DEFAULT 0;
 p_t_id integer; 
 BEGIN
 
  if arg_tree_name is null then
 		RAISE NOTICE 'arg_tree_name IS MANTATORY';
 		return ;
   end if; 
   
   
   p_t_id := tree_util.get_tree_id(arg_tree_name);
   if p_t_id is null then
   	RAISE NOTICE 'tree with name % not exists', arg_tree_name;
	return;
   end if;
 
   FOR rec IN
      SELECT * FROM
	  (SELECT 
	  n.*, 
      (tree.get_parent(n.id)) as parent
      FROM tree_util.tree_node n  
      WHERE
      n.level is not null and
      n.level <= arg_max_level and n.tree_id = p_t_id ) as s
      order by level, (parent).weight, (parent).internal_weight, weight, internal_weight
    LOOP
    	i := i+1;
    	rep := tree_util.get_traversal_node((rec.id,  rec.node_name, rec.node_data_id, rec.weight, rec.internal_weight, rec.level, rec.parent_id )::tree.tree_node,i);
		return next rep;    	
    END LOOP;
    return;
   
 END
 $$ STABLE LANGUAGE plpgsql;
 
 
 
 
  CREATE OR REPLACE FUNCTION  tree.traverse_tree_level(arg_tree_name varchar, arg_max_level integer) RETURNS  setof  tree.traversal_node AS $$
	SELECT * FROM tree_util.traverse_tree_level($1,$2);
  $$ STABLE LANGUAGE sql;
 
CREATE OR REPLACE FUNCTION  tree.traverse_tree_level(arg_tree_name varchar) RETURNS  setof  tree.traversal_node AS $$
	SELECT * FROM tree_util.traverse_tree_level($1);
  $$ STABLE LANGUAGE sql;
  
---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 


CREATE OR REPLACE FUNCTION tree_util.traverse_group(arg_node tree.traversal_group, arg_max_level integer) RETURNS  setof  tree.traversal_group AS $$
DECLARE
	id1 tree.tree_node ;
	id2 tree.traversal_group;
	rep tree.traversal_group;
	rec record;
	p_tag_flag boolean;
begin
    return next arg_node;
  
    SELECT into rec count(*) FROM tree.get_childs(arg_node.node_id);
    IF rec.count > 0 THEN
      p_tag_flag := true;
      rep := tree_util.get_traversal_group(null,null,1,null);
      return next rep;
    ELSE 
    	p_tag_flag := false;
    END IF;
    
    FOR id1 IN
    SELECT * FROM tree.get_childs(arg_node.node_id)
    LOOP
    	IF arg_max_level < 0 OR id1.level <= arg_max_level THEN
         FOR id2 IN
         SELECT * FROM tree_util.traverse_group( tree_util.get_traversal_group(id1,null,null,null) ,arg_max_level)
         LOOP
            return next id2;
         END LOOP;
        END IF;
    END LOOP;
    
    IF p_tag_flag THEN
    	rep := tree_util.get_traversal_group(null,null,2,null);
   		return next rep;
    END IF;
END
$$ STABLE LANGUAGE plpgsql;


---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 

CREATE OR REPLACE FUNCTION tree_util.traverse_group(arg_node_id bigint, arg_max_level integer, get_root_group boolean) RETURNS  setof  tree.traversal_group AS $$
DECLARE
	rec tree.traversal_group ;
	i1 integer DEFAULT 0;
	i2 integer DEFAULT 0;
begin
 
 
if get_root_group = true then
 	i1 := i1 + 1;
 	rec := tree_util.get_traversal_group(null,i1,1,null);
    return next rec;
end if;
 
 FOR rec IN
    SELECT * FROM tree_util.traverse_group(tree_util.get_traversal_group(tree.get_node(arg_node_id),null,null,null), arg_max_level)
    LOOP
		if (get_root_group = true or rec.level > 0 or rec.level is null) then
			i1 := i1 + 1;
    		rec.traversal_index := i1;
			if rec.node_id is not null then
    			i2 := i2 + 1;
    			rec.node_traversal_index := i2;
    		else 
    			rec.node_traversal_index := null;
			end if;
		
			return next rec;
		end if;
    END LOOP;
    
    if get_root_group = true then
 		i1 := i1 + 1;
 		rec := tree_util.get_traversal_group(null,i1,2,null);
   		return next rec;
 	end if;

	return;
	
END
$$ STABLE LANGUAGE plpgsql;


---------------------------------------------------------------------------
--
---------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION tree.traverse_group(arg_node_id bigint, arg_max_level integer, get_root_group boolean) RETURNS  setof  tree.traversal_group AS $$
    SELECT * FROM tree_util.traverse_group($1,$2,$3);
$$ STABLE LANGUAGE sql;


CREATE OR REPLACE FUNCTION tree.traverse_group(arg_node_id bigint, get_root_group boolean) RETURNS  setof  tree.traversal_group AS $$
    SELECT * FROM tree.traverse_group($1,-1,$2)
$$ STABLE LANGUAGE sql;


CREATE OR REPLACE FUNCTION tree.traverse_group(arg_tree_name varchar, get_root_group boolean) RETURNS  setof  tree.traversal_group AS $$
    SELECT * FROM tree.traverse_group(tree.get_root_id($1),-1,$2)
$$ STABLE LANGUAGE sql;


CREATE OR REPLACE FUNCTION tree.traverse_group(arg_tree_name varchar, arg_max_level integer, get_root_group boolean) RETURNS  setof  tree.traversal_group AS $$
    SELECT * FROM tree.traverse_group(tree.get_root_id($1),$2,$3)
$$ STABLE LANGUAGE sql;

